/* ****************************************************** */
/*    Wilson fermion solver                               */
/*                                                        */
/*    This program is written in C-like manner, but       */
/*   using C++ functions.  Please compile by g++, or      */
/*   other C++ compiler.  [18 Apr 2009 HM]                */
/*                                                        */
/*                     Copyright(c) Hideo Matsufuru 2010  */
/* ****************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "lattsize.h"

#include "field.h"

#include "solv_wilson_ocl.h"

#include "linear_algebra_c.h"

#define MAX_SOURCE_SIZE (0x1000000)


/* ****************************************************** */
void Solv_Wilson::solve_CG(double enorm, int &nconv, double &diff,
                       Field &fxq, Field &fu, Field &fb){

  static double xq[Nvst], b[Nvst], u[Ndf*Nst*4];

  static double x[Nvst], s[Nvst], r[Nvst], p[Nvst], v[Nvst];

  int niter = 1000;

  //OpenCL
  set_opencl();

  set_f_to_a(fu,u);
  set_f_to_a(fb,b);


  equate_c(Nvst, s, b);
  double snorm, sr;
  normv_c(Nvst, sr, s);
  snorm = 1.0/sr;

  double rr, rrp;
  nconv = -1;

  solve_CG_init(rrp, rr, u, x, r, s, p);
    printf("  init: %22.15e\n", rr*snorm);

  for(int iter=0; iter < niter; iter++){

    solve_CG_step(rrp, rr, u, x, r, s, p);

       printf("%6d  %22.15e\n", iter, rr*snorm);

    if(rr*snorm < enorm){
      nconv = iter;
      break;
    }

  }

  if(nconv == -1){
    printf(" not converged\n");
  }

  //  printf("converged:\n");
  //  printf("  nconv = %d\n", nconv);

  equate_c(Nvst,xq,x);

  opr_DdagD(r,u,x);
  selfadd_c(Nvst,r,b,-1.0);
  normv_c(Nvst,diff,r);


  set_f_from_a(fxq,xq);


  //OpenCL
  fin_opencl();

}
/* ****************************************************** */
void Solv_Wilson::solve_CG_init(double &rrp, double &rr, double *u,
     double *x, double *r, double *s, double *p){

  equate_c(Nvst,r,s);
  equate_c(Nvst,x,s);

  opr_DdagD(s,u,x);

  selfadd_c(Nvst,r,s,-1.0);
  equate_c(Nvst,p,r);

  double rrt;
  normv_c(Nvst,rrt,r);
  rr = rrt;
  rrp = rr;

}
/* ****************************************************** */
void Solv_Wilson::solve_CG_step(double &rrp, double &rr, double *u,
     double *x, double *r, double *s, double *p){

  double cr, bk, pap;

  opr_DdagD(s,u,p);

  vecprd_c(Nvst,pap,s,p);
  cr = rrp/pap;

  selfadd_c(Nvst,x,p,cr);

  selfadd_c(Nvst,r,s,-cr);
  double rrt;
  normv_c(Nvst,rrt,r);
  rr=rrt;
  bk = rr/rrp;

  selfsprd_c(Nvst,p,bk);
  selfadd_c(Nvst,p,r,1.0);

  rrp = rr;

}

/* ****************************************************** */
void Solv_Wilson::set_f_from_a(Field &fv, double *v){

  int n = fv.size();

  for(int i=0; i<n; i++){
    fv.e(i) = v[i];
  }
}
/* ****************************************************** */
void Solv_Wilson::set_f_to_a(Field &fv, double *v){

  int n = fv.size();

  for(int i=0; i<n; i++){
    v[i] = fv.e(i);
  }
}

/* ****************************************************** */
void Solv_Wilson::set_opencl(){


  /*  get information of device platform  */
  ret = clGetPlatformIDs(1, &platform_id, &ret_num_platforms);
  printf(" number of available platforms: %d\n", ret_num_platforms);

  ret = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_DEFAULT, 1,
		       &device_id, &ret_num_devices);
  printf(" number of available devices: %d\n", ret_num_devices);

  /* create OpenCL context  */
  context = clCreateContext(NULL, 1, &device_id, NULL, NULL, &ret);

  /* create command queue  */
  command_queue = clCreateCommandQueue(context, device_id, 0, &ret);


  /* create kernel program  */
  FILE *fp;
  char *source_str;
  size_t source_size;

  char filename[] = "./mult_float.cl";

  fp = fopen(filename, "r");
  if(!fp){
      fprintf(stderr, "Failed to load kernel.\n");
      exit(1);
  }
  source_str = (char*)malloc(MAX_SOURCE_SIZE);
  source_size = fread(source_str, 1, MAX_SOURCE_SIZE, fp);
  fclose(fp);

  /* create kernel program from source file  */
  program = clCreateProgramWithSource(context, 1,
      (const char**)&source_str, (const size_t *)&source_size, &ret);

  /* build kernel program  */
  ret = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);

  if(ret!=CL_SUCCESS){
      printf(" Program build failed: %d \n", ret);
      exit(1);
  }

  free(source_str);


  /* create OpenCL kernel  */
  clmult_all = clCreateKernel(program, "mult_all", &ret);

}
/* ****************************************************** */
void Solv_Wilson::fin_opencl(){

// release OpenCL items

  ret = clReleaseKernel(clmult_all);

  ret = clFlush(command_queue);
  ret = clFinish(command_queue);
  ret = clReleaseProgram(program);
  ret = clReleaseCommandQueue(command_queue);
  ret = clReleaseContext(context);


}
/* ****************************************************** */
void Solv_Wilson::opr_DdagD(double *v, double *u, double *w){

  static double v2[Nvst];

  opr_H(v2,u,w );
  opr_H(v ,u,v2);

}
/* ****************************************************** */
void Solv_Wilson::opr_D(double *v, double *u, double *w){

/*
  setconst_c(Nvst, v, 0.0);

  mult_xp(v,u,w);
  mult_xm(v,u,w);
  mult_yp(v,u,w);
  mult_ym(v,u,w);
  mult_zp(v,u,w);
  mult_zm(v,u,w);
  mult_tp(v,u,w);
  mult_tm(v,u,w);

  spx_c(Nvst,v,w,-CKs);
*/

}
/* ****************************************************** */
void Solv_Wilson::opr_H(double *v, double *u, double *w){

  cl_mem            Vmobj = NULL;
  cl_mem            Wmobj = NULL;
  cl_mem            Umobj = NULL;

  real_t vf[Nvst];
  real_t wf[Nvst];
  real_t uf[Ndf*Nst*4];

//  setconst_c(Nvst, v, 0.0);

  for(int i=0; i<Nvst; i++){
      vf[i] = 0.0;
      wf[i] = (float)w[i];
  }
  for(int i=0; i<(Ndf*Nst*4); i++){
      uf[i] = (float)u[i];
  }

  int Nmemtrf = Nvst * sizeof(real_t);
  int Nmemtrg = Ndf*Nst*4 * sizeof(real_t);

  real_t CKs2 = (real_t) CKs;

  Vmobj = clCreateBuffer(context, CL_MEM_READ_WRITE,
			 Nmemtrf, NULL, &ret);

  Wmobj = clCreateBuffer(context, CL_MEM_READ_WRITE,
			 Nmemtrf, NULL, &ret);

  Umobj = clCreateBuffer(context, CL_MEM_READ_WRITE,
			 Nmemtrg, NULL, &ret);

  ret = clEnqueueWriteBuffer(command_queue, Vmobj, CL_TRUE,
		     0, Nmemtrf, vf, 0, NULL, NULL);
  ret = clEnqueueWriteBuffer(command_queue, Wmobj, CL_TRUE,
		     0, Nmemtrf, wf, 0, NULL, NULL);

  ret = clEnqueueWriteBuffer(command_queue, Umobj, CL_TRUE,
		     0, Nmemtrg, uf, 0, NULL, NULL);

  size_t global_item_size = Nst;
  size_t local_item_size = 1;

  // mult_all
  ret = clSetKernelArg(clmult_all, 0, sizeof(cl_mem), (void *)&Vmobj);
  ret = clSetKernelArg(clmult_all, 1, sizeof(cl_mem), (void *)&Umobj);
  ret = clSetKernelArg(clmult_all, 2, sizeof(cl_mem), (void *)&Wmobj);
  ret = clSetKernelArg(clmult_all, 3, sizeof(float), (void *)&CKs2);

  ret = clEnqueueNDRangeKernel(command_queue, clmult_all, 1, NULL,
	       &global_item_size, &local_item_size, 0, NULL, NULL);


  /* read result from memory buffer  */
  ret = clEnqueueReadBuffer(command_queue, Vmobj, CL_TRUE, 0,
			    Nmemtrf, vf, 0, NULL, NULL);

/* For float  */

  for(int i=0;i<Nvst;i++){
      v[i] = (double)vf[i];
      w[i] = (double)wf[i];
  }

  /* finishing  */
  ret = clReleaseMemObject(Vmobj);
  ret = clReleaseMemObject(Umobj);
  ret = clReleaseMemObject(Wmobj);

}

/* ****************************************************** */
